const std = @import("std");
const Step = std.build.Step;
const assert = std.debug.assert;
const mem = std.mem;

const Scanner = @import("zig-wayland").Scanner;

/// While a zelbar release is in development, this string should contain
/// the version in development with the "-dev" suffix.  When a release is
/// tagged, the "-dev" suffix should be removed for the commit that gets tagged.
/// Directly after the tagged commit, the version should be bumped and the "-dev"
/// suffix added.
const version = "1.1.0";

pub fn build(b: *std.Build) !void {
    const target = b.standardTargetOptions(.{});
    const optimize = b.standardOptimizeOption(.{});

    const pie = b.option(bool, "pie", "Build a Position Independent Executable") orelse false;

    const full_version = blk: {
        if (mem.endsWith(u8, version, "-dev")) {
            var ret: u8 = undefined;

            const git_describe_long = b.runAllowFail(
                &[_][]const u8{ "git", "-C", b.build_root.path orelse ".", "describe", "--long" },
                &ret,
                .Inherit,
            ) catch break :blk version;

            var it = mem.split(u8, mem.trim(u8, git_describe_long, &std.ascii.whitespace), "-");
            _ = it.next().?; // previous tag
            const commit_count = it.next().?;
            const commit_hash = it.next().?;
            assert(it.next() == null);
            assert(commit_hash[0] == 'g');

            // Follow semantic versioning, e.g. 0.2.0-dev.42+d1cf95b
            break :blk try std.fmt.allocPrintZ(b.allocator, version ++ ".{s}+{s}", .{
                commit_count,
                commit_hash[1..],
            });
        } else {
            break :blk version;
        }
    };

    const options = b.addOptions();
    options.addOption([]const u8, "version", full_version);

    const scanner = Scanner.create(b, .{});

    scanner.addSystemProtocol("stable/xdg-shell/xdg-shell.xml");
    scanner.addSystemProtocol("staging/cursor-shape/cursor-shape-v1.xml");
    scanner.addSystemProtocol("unstable/tablet/tablet-unstable-v2.xml"); // Needed for cursor-shape
    scanner.addCustomProtocol("protocol/wlr-layer-shell-unstable-v1.xml");

    scanner.generate("wl_compositor", 4);
    scanner.generate("wl_shm", 1);
    scanner.generate("wl_output", 4);
    scanner.generate("wl_seat", 5);
    scanner.generate("zwlr_layer_shell_v1", 3);
    scanner.generate("wp_cursor_shape_manager_v1", 1);

    const wayland = b.createModule(.{ .root_source_file = scanner.result });
    const flags = b.createModule(.{ .root_source_file = b.path("common/flags.zig") });
    const pixman = b.dependency("zig-pixman", .{}).module("pixman");
    const fcft = b.dependency("zig-fcft", .{}).module("fcft");

    const zelbar = b.addExecutable(.{
        .name = "zelbar",
        .root_source_file = b.path("src/main.zig"),
        .target = target,
        .optimize = optimize,
    });

    zelbar.root_module.addOptions("build_options", options);

    zelbar.linkLibC();

    zelbar.root_module.addImport("wayland", wayland);
    zelbar.linkSystemLibrary("wayland-client");
    zelbar.linkSystemLibrary("wayland-cursor");

    zelbar.root_module.addImport("flags", flags);

    zelbar.root_module.addImport("pixman", pixman);
    zelbar.linkSystemLibrary("pixman-1");

    zelbar.root_module.addImport("fcft", fcft);
    zelbar.linkSystemLibrary("fcft");

    // TODO: remove when https://github.com/ziglang/zig/issues/131 is implemented
    scanner.addCSource(zelbar);

    zelbar.pie = pie;

    b.installArtifact(zelbar);
    b.installFile("doc/zelbar.1", "share/man/man1/zelbar.1");

    {
        const tests = b.addTest(.{
            .root_source_file = b.path("src/test_main.zig"),
            .target = target,
            .optimize = optimize,
        });
        tests.root_module.addImport("wayland", wayland);
        tests.root_module.addImport("pixman", pixman);
        tests.root_module.addImport("fcft", fcft);
        tests.linkSystemLibrary("pixman-1");
        tests.linkSystemLibrary("fcft");
        const test_step = b.step("test", "Run all tests");
        test_step.dependOn(&b.addRunArtifact(tests).step);
    }
}
