// This file is part of zelbar
//
// Copyright (C) 2022 Hugo Machet
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

const std = @import("std");
const assert = std.debug.assert;
const builtin = @import("builtin");
const io = std.io;
const mem = std.mem;
const posix = std.posix;

const pixman = @import("pixman");
const wayland = @import("wayland");
const wl = wayland.client.wl;
const wp = wayland.client.wp;
const zwlr = wayland.client.zwlr;

const Bar = @import("Bar.zig");
const ctx = &@import("main.zig").context;
const Backend = @This();

const log = std.log.scoped(.backend);

const Output = struct {
    wl_output: *wl.Output,
    wl_name: u32,

    name: ?[]const u8 = null,
    // TODO: scale

    configured: bool = false,

    fn init(output: *Output, wl_output: *wl.Output, wl_name: u32) !void {
        output.* = .{
            .wl_output = wl_output,
            .wl_name = wl_name,
        };

        output.wl_output.setListener(*Output, output_listener, output);
    }

    fn destroy(output: *Output) void {
        output.wl_output.release();
        if (output.name) |name| ctx.gpa.free(name);

        const node: *std.SinglyLinkedList(Output).Node = @fieldParentPtr("data", output);
        ctx.backend.outputs.remove(node);
        ctx.gpa.destroy(node);
    }

    fn output_listener(_: *wl.Output, event: wl.Output.Event, output: *Output) void {
        switch (event) {
            .done => {
                log.info("output {?s} configured", .{output.name});
                output.configured = true;
            },
            .name => |ev| {
                output.name = ctx.gpa.dupe(u8, mem.span(ev.name)) catch |err| switch (err) {
                    error.OutOfMemory => {
                        log.err("out of memory", .{});
                        return;
                    },
                };
            },
            else => {},
        }
    }
};

// Taken and adapted from https://git.sr.ht/~leon_plickat/wayprompt, same license.
const HotSpot = struct {
    action: []const u8,
    x: u31,
    y: u31,
    width: u31,
    height: u31,

    fn contains_point(hotspot: HotSpot, x: u31, y: u31) bool {
        return x >= hotspot.x and x <= hotspot.x +| hotspot.width and
            y >= hotspot.y and y <= hotspot.y +| hotspot.height;
    }

    /// Write action in STDOUT.
    fn act(hotspot: HotSpot) !void {
        var bw = io.bufferedWriter(io.getStdOut().writer());
        const writer = bw.writer();
        try writer.writeAll(hotspot.action);
        try writer.writeByte('\n');
        try bw.flush();
    }
};

// Pointers event code have been taken and adapted from
// https://git.sr.ht/~leon_plickat/wayprompt, same license.
const Seat = struct {
    wl_seat: *wl.Seat,
    id: u32,

    // Pointer related objects.
    wl_pointer: ?*wl.Pointer = null,
    cursor_shape_device: ?*wp.CursorShapeDeviceV1 = null,
    pointer_x: u31 = 0,
    pointer_y: u31 = 0,
    cursor_shape: ?wp.CursorShapeDeviceV1.Shape = null,
    cursor_theme: ?*wl.CursorTheme = null,
    cursor_surface: ?*wl.Surface = null,
    last_enter_serial: u32 = undefined,
    press_hotspot: ?*HotSpot = null,

    fn init(seat: *Seat, wl_seat: *wl.Seat, id: u32) !void {
        seat.* = .{
            .wl_seat = wl_seat,
            .id = id,
        };

        seat.wl_seat.setListener(*Seat, seat_listener, seat);
    }

    fn destroy(seat: *Seat) void {
        seat.release_pointer();
        seat.wl_seat.release();

        const node: *std.SinglyLinkedList(Seat).Node = @fieldParentPtr("data", seat);
        ctx.backend.seats.remove(node);
        ctx.gpa.destroy(node);
    }

    fn seat_listener(_: *wl.Seat, event: wl.Seat.Event, seat: *Seat) void {
        switch (event) {
            .capabilities => |ev| {
                if (ev.capabilities.pointer) {
                    seat.bind_pointer() catch {};
                } else {
                    seat.release_pointer();
                }
            },
            .name => {},
        }
    }

    fn bind_pointer(seat: *Seat) !void {
        if (seat.wl_pointer != null) return;
        assert(seat.cursor_shape == null);
        seat.wl_pointer = try seat.wl_seat.getPointer();
        seat.wl_pointer.?.setListener(*Seat, pointer_listener, seat);
        if (ctx.backend.cursor_shape_mgr) |csm| {
            seat.cursor_shape_device = try csm.getPointer(seat.wl_pointer.?);
        }
    }

    fn release_pointer(seat: *Seat) void {
        seat.cursor_shape = null;
        seat.press_hotspot = null;
        if (seat.cursor_theme) |t| {
            t.destroy();
            seat.cursor_theme = null;
        }
        if (seat.cursor_surface) |s| {
            s.destroy();
            seat.cursor_surface = null;
        }
        if (seat.wl_pointer) |p| {
            p.release();
            seat.wl_pointer = null;
        }
        if (seat.cursor_shape_device) |csd| {
            csd.destroy();
            seat.cursor_shape_device = null;
        }
    }

    fn pointer_listener(_: *wl.Pointer, event: wl.Pointer.Event, seat: *Seat) void {
        switch (event) {
            .enter => |ev| seat.update_pointer(ev.surface_x, ev.surface_y, ev.serial),
            .leave => seat.cursor_shape = null,
            .motion => |ev| seat.update_pointer(ev.surface_x, ev.surface_y, null),
            .button => |ev| {
                // Only activating a button on release is the better UX, IMO.
                switch (ev.state) {
                    .pressed => {
                        seat.press_hotspot = ctx.backend.surface.?.hotspot_from_point(seat.pointer_x, seat.pointer_y);
                    },
                    .released => {
                        if (seat.press_hotspot == null) return;
                        if (ctx.backend.surface.?.hotspot_from_point(seat.pointer_x, seat.pointer_y)) |hs| {
                            if (hs == seat.press_hotspot.?) hs.act() catch {};
                        }
                        seat.press_hotspot = null;
                    },
                    else => {},
                }
            },
            else => {},
        }
    }

    fn update_pointer(seat: *Seat, x: wl.Fixed, y: wl.Fixed, serial: ?u32) void {
        const X = x.toInt();
        seat.pointer_x = if (X > 0) @intCast(X) else 0;

        const Y = y.toInt();
        seat.pointer_y = if (Y > 0) @intCast(Y) else 0;

        if (serial) |s| seat.last_enter_serial = s;

        // Sanity check.
        assert(seat.wl_pointer != null);
        assert(ctx.backend.surface != null);

        // Cursor errors shall not be fatal. It's fairly expectable for
        // something to go wrong there and it's not exactly vital to our
        // operation here, so we can roll without setting the cursor.
        if (ctx.backend.surface.?.hotspot_from_point(seat.pointer_x, seat.pointer_y) != null) {
            seat.set_cursor(.pointer) catch {};
        } else {
            seat.set_cursor(.default) catch {};
        }
    }

    fn set_cursor(seat: *Seat, shape: wp.CursorShapeDeviceV1.Shape) !void {
        if (seat.cursor_shape) |cs| {
            if (cs == shape) return;
        }
        seat.cursor_shape = shape;

        // If the Wayland server supports the CursorShapeV1 protocol, use that
        // to set the cursor. Otherwise we need to go through wl-cursor.
        if (seat.cursor_shape_device) |csd| {
            csd.setShape(seat.last_enter_serial, shape);
        } else {
            try seat.set_cursor_surface(shape);
        }
    }

    fn set_cursor_surface(seat: *Seat, shape: wp.CursorShapeDeviceV1.Shape) !void {
        const name = @tagName(shape);
        const cursor_size = 24;

        if (seat.cursor_theme == null) {
            seat.cursor_theme = try wl.CursorTheme.load(null, cursor_size, ctx.backend.shm.?);
        }
        errdefer {
            seat.cursor_theme.?.destroy();
            seat.cursor_theme = null;
        }

        // These just point back to the CursorTheme, no need to keep them.
        const wl_cursor = seat.cursor_theme.?.getCursor(name) orelse return error.NoCursor;
        const cursor_image = wl_cursor.images[0];
        const wl_buffer = try cursor_image.getBuffer();

        if (seat.cursor_surface == null) {
            seat.cursor_surface = try ctx.backend.compositor.?.createSurface();
        }
        errdefer {
            seat.cursor_surface.?.destroy();
            seat.cursor_surface = null;
        }

        seat.cursor_surface.?.attach(wl_buffer, 0, 0);
        seat.cursor_surface.?.damageBuffer(0, 0, std.math.maxInt(i31), std.math.maxInt(u31));
        seat.cursor_surface.?.commit();

        seat.wl_pointer.?.setCursor(
            seat.last_enter_serial,
            seat.cursor_surface.?,
            @intCast(cursor_image.hotspot_x),
            @intCast(cursor_image.hotspot_y),
        );
    }
};

const Surface = struct {
    wl_surface: *wl.Surface = undefined,
    layer_surface: *zwlr.LayerSurfaceV1 = undefined,

    hotspots: std.ArrayListUnmanaged(HotSpot) = .{},

    frame_callback: ?*wl.Callback = null,
    /// True if we requested a frame callback.
    frame_pending: bool = false,
    // True if we need to redraw.
    dirty: bool = false,

    width: u31 = 0,
    height: u31 = 0,

    configured: bool = false,

    fn deinit(surface: *Surface) void {
        surface.configured = false;
        surface.hotspots.deinit(ctx.gpa);
        if (surface.frame_callback) |cb| cb.destroy();
        surface.layer_surface.destroy();
        surface.wl_surface.destroy();
    }

    fn send_frame(surface: *Surface) !void {
        // The surface is not initialized yet.
        if (!surface.configured) {
            const wl_surface = try ctx.backend.compositor.?.createSurface();
            const layer_surface = try ctx.backend.layer_shell.?.getLayerSurface(
                wl_surface,
                ctx.backend.req_output.wl_output,
                switch (ctx.bar.config.layer) {
                    .bottom => .bottom,
                    .top => .top,
                    .overlay => .overlay,
                },
                "panel",
            );

            surface.* = .{
                .wl_surface = wl_surface,
                .layer_surface = layer_surface,
            };
            surface.layer_surface.setListener(*Surface, layer_surface_listener, surface);
        }

        // Configure the layer_surface and resize it if needed.
        if (surface.height != ctx.bar.config.height) {
            const cfg = ctx.bar.config;
            surface.layer_surface.setSize(0, cfg.height);
            if (cfg.layer == .overlay) {
                surface.layer_surface.setExclusiveZone(0);
            } else {
                surface.layer_surface.setExclusiveZone(cfg.height);
            }
            switch (cfg.location) {
                .top => {
                    surface.layer_surface.setAnchor(.{ .top = true, .right = true, .bottom = false, .left = true });
                },
                .bottom => {
                    surface.layer_surface.setAnchor(.{ .top = false, .right = true, .bottom = true, .left = true });
                },
            }
            surface.layer_surface.setMargin(cfg.margins.top, cfg.margins.right, cfg.margins.bottom, cfg.margins.left);

            // We need to commit the empty surface first so we can receive a configure
            // event with width and height requested for our surface.
            surface.wl_surface.commit();

            return;
        }

        // If we made it till here we can finally draw on your surface.
        // Get a new buffer not busy.
        const buffer = try ctx.backend.pool.next_buffer(surface.width, surface.height);
        assert(!buffer.busy);

        try ctx.bar.render(buffer, @intCast(surface.width), @intCast(surface.height));

        // Attach the buffer to the surface.
        surface.wl_surface.attach(buffer.wl_buffer, 0, 0);
        surface.wl_surface.damageBuffer(0, 0, buffer.width, buffer.height);
        buffer.busy = true;

        // Schedule a frame in case the surface become dirty again.
        surface.schedule_frame_and_commit();
        surface.dirty = false;
    }

    pub fn set_dirty(surface: *Surface) void {
        if (surface.dirty) return;
        surface.dirty = true;
        surface.schedule_frame_and_commit();
    }

    fn schedule_frame_and_commit(surface: *Surface) void {
        if (surface.frame_pending) return;

        // Request a new callback for the next frame.
        surface.frame_callback = surface.wl_surface.frame() catch return;
        surface.frame_callback.?.setListener(*Surface, frame_callback_listener, surface);
        surface.wl_surface.commit();
        surface.frame_pending = true;
    }

    fn hotspot_from_point(surface: *Surface, x: u31, y: u31) ?*HotSpot {
        for (surface.hotspots.items) |*hs| {
            if (hs.contains_point(x, y)) return hs;
        }
        return null;
    }

    fn layer_surface_listener(
        _: *zwlr.LayerSurfaceV1,
        event: zwlr.LayerSurfaceV1.Event,
        surface: *Surface,
    ) void {
        switch (event) {
            .configure => |ev| {
                surface.layer_surface.ackConfigure(ev.serial);

                const w: u31 = @truncate(ev.width);
                const h: u31 = @truncate(ev.height);

                if (surface.configured and surface.width == w and surface.height == h) {
                    surface.wl_surface.commit();
                    return;
                }

                surface.configured = true;
                surface.width = w;
                surface.height = h;

                surface.send_frame() catch return;
            },
            .closed => {
                surface.deinit();
                ctx.gpa.destroy(surface);
                ctx.backend.surface = null;
            },
        }
    }

    fn frame_callback_listener(_: *wl.Callback, event: wl.Callback.Event, surface: *Surface) void {
        switch (event) {
            .done => {
                // Destroy the now used callback.
                if (surface.frame_callback) |cb| cb.destroy();
                surface.frame_callback = null;
                surface.frame_pending = false;

                if (surface.dirty) surface.send_frame() catch return;
            },
        }
    }
};

// Taken and adapted from https://git.sr.ht/~leon_plickat/wayprompt, same license.
const BufferPool = struct {
    /// The amount of buffers per surface we consider the reasonable upper limit.
    /// Some compositors sometimes tripple-buffer, so three seems to be ok.
    /// Note that we can absolutely work with higher buffer numbers if needed,
    /// however we consider that to be an anomaly and therefore do not want to
    /// keep all those extra buffers around if we can avoid it, as to not have
    /// unecessary memory overhead.
    const max_buffer_multiplicity = 3;

    /// The buffers. This is a linked list and not an array list, because we
    /// need stable pointers for the listener of the wl_buffer object.
    buffers: std.TailQueue(Buffer) = .{},

    /// Deinit the buffer pool, destroying all buffers and freeing all memory.
    pub fn deinit(self: *BufferPool) void {
        var it = self.buffers.first;
        while (it) |node| {
            // We need to get the next node before destroying the current one.
            it = node.next;
            node.data.deinit();
            ctx.gpa.destroy(node);
        }
    }

    /// Get a buffer of the specified dimenisons. If possible an idle buffer is
    /// reused, otherweise a new one is created.
    pub fn next_buffer(self: *BufferPool, width: u31, height: u31) !*Buffer {
        log.debug("Next buffer: {}x{}; Total buffers: {}", .{ width, height, self.buffers.len });
        defer {
            if (self.buffers.len > max_buffer_multiplicity) {
                self.cull_buffers();
            }
        }
        if (try self.find_suitable_buffer(width, height)) |buffer| {
            return buffer;
        } else {
            return try self.new_buffer(width, height);
        }
    }

    fn find_suitable_buffer(self: *BufferPool, width: u31, height: u31) !?*Buffer {
        var it = self.buffers.first;
        var first_unbusy_buffer_node: ?*std.TailQueue(Buffer).Node = null;
        while (it) |node| : (it = node.next) {
            if (node.data.busy) continue;
            if (node.data.width == width and node.data.height == height) {
                return &node.data;
            } else {
                first_unbusy_buffer_node = node;
            }
        }

        // No buffer has matching dimensions, however we do have an unbusy
        // buffer which we can just re-init.
        if (first_unbusy_buffer_node) |node| {
            node.data.deinit();
            try node.data.init(width, height);
            return &node.data;
        }

        return null;
    }

    fn new_buffer(self: *BufferPool, width: u31, height: u31) !*Buffer {
        log.debug("New buffer: {}x{}", .{ width, height });
        const node = try ctx.gpa.create(std.TailQueue(Buffer).Node);
        errdefer ctx.gpa.destroy(node);
        try node.data.init(width, height);
        self.buffers.append(node);
        return &node.data;
    }

    fn cull_buffers(self: *BufferPool) void {
        log.debug("Culling buffers.", .{});
        var overhead = self.buffers.len - max_buffer_multiplicity;
        var it = self.buffers.first;
        while (it) |node| {
            if (overhead == 0) break;
            // We need to get the next node before destroying the current one.
            it = node.next;
            if (!node.data.busy) {
                node.data.deinit();
                self.buffers.remove(node);
                ctx.gpa.destroy(node);
                overhead -= 1;
            }
        }
        log.debug(" -> new buffer count: {}", .{self.buffers.len});
    }
};

pub const Buffer = struct {
    wl_buffer: ?*wl.Buffer = null,
    data: ?[]align(std.mem.page_size) u8 = null,
    pixman_image: ?*pixman.Image = null,

    width: u31 = 0,
    height: u31 = 0,

    busy: bool = false,

    fn init(buffer: *Buffer, width: u31, height: u31) !void {
        const stride = width << 2;
        const size: u31 = stride * height;

        // Open a memory backed file.
        const fd = switch (builtin.target.os.tag) {
            .linux, .freebsd => try posix.memfd_createZ("zelbar-shm-buffer-pool", std.os.linux.MFD.CLOEXEC),
            else => @compileError("target OS not supported"),
        };
        defer posix.close(fd);
        try posix.ftruncate(fd, size);

        // mmap the memory file, to be used by the pximan image.
        const data = mem.bytesAsSlice(
            u8,
            try posix.mmap(null, size, posix.PROT.READ | posix.PROT.WRITE, .{ .TYPE = .SHARED }, fd, 0),
        );
        errdefer posix.munmap(data);

        // Create a Wayland shm buffer for the same memory file.
        const shm = ctx.backend.shm orelse return error.NoShm;
        const pool = try shm.createPool(fd, size);
        defer pool.destroy();

        const wl_buffer = try pool.createBuffer(0, width, height, stride, .argb8888);
        errdefer wl_buffer.destroy();
        wl_buffer.setListener(*Buffer, buffer_listener, buffer);

        // Create the pixman image.
        const pixman_image = pixman.Image.createBitsNoClear(
            .a8r8g8b8,
            @intCast(width),
            @intCast(height),
            @ptrCast(data),
            @intCast(stride),
        );
        errdefer _ = pixman_image.unref();

        // The pixman image and the Wayland buffer now share the same memory.
        buffer.* = .{
            .wl_buffer = wl_buffer,
            .data = data,
            .pixman_image = pixman_image,
            .width = width,
            .height = height,
        };
    }

    fn deinit(buffer: Buffer) void {
        if (buffer.pixman_image) |pix| _ = pix.unref();
        if (buffer.wl_buffer) |wlb| wlb.destroy();
        if (buffer.data) |data| posix.munmap(data);
    }

    fn buffer_listener(_: *wl.Buffer, event: wl.Buffer.Event, buffer: *Buffer) void {
        switch (event) {
            .release => {
                assert(buffer.busy);
                buffer.busy = false;
            },
        }
    }
};

display: *wl.Display = undefined,
registry: *wl.Registry = undefined,

compositor: ?*wl.Compositor = null,
shm: ?*wl.Shm = null,
layer_shell: ?*zwlr.LayerShellV1 = null,
cursor_shape_mgr: ?*wp.CursorShapeManagerV1 = null,

outputs: std.SinglyLinkedList(Output) = .{},
seats: std.SinglyLinkedList(Seat) = .{},

/// Requested output by the user to display the bar on.
// TODO: What if the req_output is destroyed
req_output: *Output = undefined,
surface: ?Surface = null,
pool: BufferPool = .{},

pub fn init(backend: *Backend) !bool {
    backend.* = .{
        .display = wl.Display.connect(null) catch {
            ctx.fatal(.backend, "failed to connect to Wayland compositor", .{});
        },
        .registry = try backend.display.getRegistry(),
        .surface = Surface{},
    };

    backend.registry.setListener(*Backend, registry_listener, backend);

    const errno = backend.display.roundtrip();
    if (errno != .SUCCESS) {
        ctx.fatal(.backend, "initial roundtrip failed: E{s}", .{@tagName(errno)});
    }

    if (backend.compositor == null) fatal_not_advertised(wl.Compositor);
    if (backend.shm == null) fatal_not_advertised(wl.Shm);
    if (backend.layer_shell == null) fatal_not_advertised(zwlr.LayerShellV1);

    // Roundtrip to get all outputs/seats infos.
    _ = backend.display.roundtrip();

    // Get the output requested by the user or use the first one in the outputs list.
    if (ctx.bar.config.output_name) |name| {
        var it = backend.outputs.first;
        while (it) |node| : (it = node.next) {
            const output_name = node.data.name orelse continue;
            if (mem.eql(u8, name, output_name)) {
                backend.req_output = &node.data;
                break;
            }
        } else {
            ctx.fatal(.backend, "no output with this name: {s}", .{name});
        }
    } else {
        const node = backend.outputs.first orelse return error.NoOutputDetected;
        log.info("no output requested, using the first in the list({?s})", .{node.data.name});
        backend.req_output = &node.data;
    }

    // We finally have all the infos needed to create the surface and start rendering.
    try backend.surface.?.send_frame();

    return backend.display.roundtrip() == .SUCCESS;
}

pub fn deinit(backend: *Backend) void {
    backend.pool.deinit();

    if (backend.surface) |*s| {
        s.deinit();
        backend.surface = null;
    }

    if (backend.compositor) |cmp| cmp.destroy();
    if (backend.shm) |shm| shm.destroy();
    if (backend.layer_shell) |ls| ls.destroy();
    if (backend.cursor_shape_mgr) |csm| csm.destroy();

    while (backend.outputs.first) |node| node.data.destroy();
    while (backend.seats.first) |node| node.data.destroy();

    backend.registry.destroy();
    backend.display.disconnect();

    backend.* = undefined;
}

/// ref: https://github.com/ifreund/waylock/blob/master/src/Lock.zig#L184
/// This function does the following:
///  1. Dispatch buffered wayland events to their listener callbacks.
///  2. Prepare the wayland connection for reading.
///  3. Send all buffered wayland requests to the server.
/// After this function has been called, either wl.Display.readEvents() or
/// wl.Display.cancelRead() read must be called.
pub fn flush_wayland_and_prepare_read(backend: *Backend) void {
    while (!backend.display.prepareRead()) {
        const errno = backend.display.dispatchPending();
        if (errno != .SUCCESS) {
            ctx.fatal(.backend, "failed to dispatch pending wayland events: E{s}", .{@tagName(errno)});
        }
    }

    while (true) {
        const errno = backend.display.flush();
        switch (errno) {
            .SUCCESS => return,
            .PIPE => {
                // libwayland uses this error to indicate that the wayland server
                // closed its side of the wayland socket. We want to continue to
                // read any buffered messages from the server though as there is
                // likely a protocol error message we'd like libwayland to log.
                _ = backend.display.readEvents();
                ctx.fatal(.backend, "connection to wayland server unexpectedly terminated", .{});
            },
            .AGAIN => {
                // The socket buffer is full, so wait for it to become writable again.
                var wayland_out = [_]posix.pollfd{.{
                    .fd = backend.display.getFd(),
                    .events = posix.POLL.OUT,
                    .revents = 0,
                }};
                _ = posix.poll(&wayland_out, -1) catch |err| {
                    ctx.fatal(.backend, "poll() failed: {s}", .{@errorName(err)});
                };
                // No need to check for POLLHUP/POLLERR here, just fall
                // through to the next flush() to handle them in one place.
            },
            else => {
                ctx.fatal(.backend, "failed to flush wayland requests: E{s}", .{@tagName(errno)});
            },
        }
    }
}

fn registry_listener(registry: *wl.Registry, event: wl.Registry.Event, backend: *Backend) void {
    backend.registry_event(registry, event) catch |err| switch (err) {
        error.OutOfMemory => {
            log.err("out of memory", .{});
            return;
        },
        else => return,
    };
}

fn registry_event(backend: *Backend, registry: *wl.Registry, event: wl.Registry.Event) !void {
    switch (event) {
        .global => |ev| {
            if (mem.orderZ(u8, ev.interface, wl.Compositor.getInterface().name) == .eq) {
                // Version 4 required for wl_surface.damage_buffer.
                if (ev.version < 4) fatal_version(wl.Compositor, ev.version, 4);
                backend.compositor = try registry.bind(ev.name, wl.Compositor, 4);
            } else if (mem.orderZ(u8, ev.interface, wl.Shm.getInterface().name) == .eq) {
                backend.shm = try registry.bind(ev.name, wl.Shm, 1);
            } else if (mem.orderZ(u8, ev.interface, wl.Output.getInterface().name) == .eq) {
                // Version 4 required for wl_output.name.
                if (ev.version < 4) fatal_version(wl.Output, ev.version, 4);

                const wl_output = try registry.bind(ev.name, wl.Output, 4);
                errdefer wl_output.release();

                const node = try ctx.gpa.create(std.SinglyLinkedList(Output).Node);
                errdefer ctx.gpa.destroy(node);

                try node.data.init(wl_output, ev.name);
                backend.outputs.prepend(node);
            } else if (mem.orderZ(u8, ev.interface, wl.Seat.getInterface().name) == .eq) {
                // Version 5 required for wl_seat.release.
                if (ev.version < 5) fatal_version(wl.Seat, ev.version, 5);

                const wl_seat = try registry.bind(ev.name, wl.Seat, 5);
                errdefer wl_seat.release();

                const node = try ctx.gpa.create(std.SinglyLinkedList(Seat).Node);
                errdefer ctx.gpa.destroy(node);

                try node.data.init(wl_seat, ev.name);
                backend.seats.prepend(node);
            } else if (mem.orderZ(u8, ev.interface, zwlr.LayerShellV1.getInterface().name) == .eq) {
                // Version 3 required for layer_shell.destroy();
                if (ev.version < 3) fatal_version(zwlr.LayerShellV1, ev.version, 3);
                backend.layer_shell = try registry.bind(ev.name, zwlr.LayerShellV1, 3);
            } else if (mem.orderZ(u8, ev.interface, wp.CursorShapeManagerV1.getInterface().name) == .eq) {
                if (ev.version < 1) fatal_version(wp.CursorShapeManagerV1, ev.version, 1);
                backend.cursor_shape_mgr = try registry.bind(ev.name, wp.CursorShapeManagerV1, 1);
            }
        },
        .global_remove => |ev| {
            {
                var it = backend.outputs.first;
                while (it) |node| : (it = node.next) {
                    if (node.data.wl_name == ev.name) {
                        node.data.destroy();
                        break;
                    }
                }
            }

            {
                var it = backend.seats.first;
                while (it) |node| : (it = node.next) {
                    if (node.data.id == ev.name) {
                        node.data.destroy();
                        break;
                    }
                }
            }
        },
    }
}

fn fatal_version(comptime Global: type, found_version: u32, req_version: u32) noreturn {
    ctx.fatal(
        .backend,
        "the advertised {s} version is too old({d}). Version {d} is required",
        .{ Global.getInterface().name, found_version, req_version },
    );
}

fn fatal_not_advertised(comptime Global: type) noreturn {
    ctx.fatal(.backend, "{s} not advertised", .{Global.getInterface().name});
}
