// This file is part of zelbar
//
// Copyright (C) 2022 Hugo Machet
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

const std = @import("std");
const assert = std.debug.assert;

const fcft = @import("fcft");

const ctx = &@import("main.zig").context;
const Buffer = @import("Backend.zig").Buffer;
const data = @import("data.zig");
const decorations = @import("decorations.zig");
const Text = @import("Text.zig");
const Bar = @This();

const log = std.log.scoped(.bar);

pub const Config = struct {
    output_name: ?[]const u8 = null,

    /// Location of the bar in the output.
    location: enum { top, bottom } = .top,
    /// Layer of the shell surface.
    layer: enum { top, bottom, overlay } = .bottom,
    // Alignment on the output if width is not 0.
    alignment: enum { left, center, right } = .left,

    width: u16 = 0,
    height: u16 = 26,

    /// Margins around the Surface.
    margins: struct { top: u16, left: u16, bottom: u16, right: u16 } = .{
        .top = 0,
        .left = 0,
        .bottom = 0,
        .right = 0,
    },

    colors: std.AutoHashMapUnmanaged(enum { bg, fg, line }, []const u8) = .{},
    border: decorations.Border = .{},
    line_height: u16 = 2,

    fonts_name: []const u8 = "sans",
};

config: Config = .{},
font: *fcft.Font = undefined,

/// True if user requested a width (i.e. config.width > 0). If false the width
/// received in the layer_surface configure event is used.
req_width: bool = false,

/// Raw input received from STDIN.
input: std.ArrayListUnmanaged(u8) = .{},

pub fn new(config: Config) !Bar {
    return Bar{
        .config = config,
        .font = try Text.get_fcft_fonts(config.fonts_name),
        .req_width = if (config.width > 0) true else false,
    };
}

pub fn deinit(bar: *Bar) void {
    bar.font.destroy();
    bar.config.colors.deinit(ctx.gpa);
    bar.input.deinit(ctx.gpa);
}

/// Draw the bar on the Buffer.
/// surface_w is always the width of the output.
pub fn render(bar: *Bar, buffer: *Buffer, surface_w: u16, h: u16) !void {
    // Width where we draw our bar.
    const w = if (bar.req_width) bar.config.width else surface_w;

    bar.config.border.check_size(w, h) catch |err| switch (err) {
        error.BordersGreaterHeight => ctx.fatal(.bar, "borders greater than bar height({d})", .{h}),
        error.BordersGreaterWidth => ctx.fatal(.bar, "borders greater than bar width({d})", .{w}),
        else => return err,
    };

    // Starting position of the bar in the output.
    const x: u16 = switch (bar.config.alignment) {
        .left => 0,
        .center => surface_w / 2 - (w >> 1),
        .right => surface_w - w,
    };

    // Render the background and the borders of the bar.
    const pix = buffer.pixman_image orelse return error.PixmanImageEmpty;
    const bg = bar.config.colors.get(.bg).?;
    try decorations.draw_rectangle(pix, @intCast(x), 0, w, h, bg);
    try bar.config.border.draw(pix, @intCast(x), 0, w, h);

    // Parse and render the raw input if any.
    if (bar.input.items.len > 0) try data.expose(bar.input.items[0..], pix, x, w);
}
