// This file is part of zelbar
//
// Copyright (C) 2022 Hugo Machet
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

const std = @import("std");
const mem = std.mem;
const unicode = std.unicode;

const fcft = @import("fcft");
const pixman = @import("pixman");

const ctx = &@import("main.zig").context;
const decorations = @import("decorations.zig");
const Text = @This();

const log = std.log.scoped(.text);

const Mode = union(enum) {
    run: *const fcft.TextRun,
    glyphs: struct {
        glyphs: []*const fcft.Glyph,
        kerns: []c_long,
    },
};

mode: Mode,
/// The width the text takes.
width: i32 = 0,

pub fn new(ally: mem.Allocator, bytes: []const u8, font: *fcft.Font) !Text {
    // Convert utf-8 string to utf-32.
    const cp_count = try unicode.utf8CountCodepoints(bytes);
    const codepoints = try ally.alloc(u32, cp_count);
    defer ally.free(codepoints);
    {
        var i: usize = 0;
        var it = (try unicode.Utf8View.init(bytes)).iterator();
        while (it.nextCodepoint()) |cp| : (i += 1) {
            codepoints[i] = cp;
        }
    }

    // Use fcft text run if possible.
    if ((fcft.capabilities() & fcft.Capabilities.text_run_shaping) != 0) {
        const run = try font.rasterizeTextRunUtf32(codepoints, .none);

        var text_w: i32 = 0;
        var i: usize = 0;
        while (i < run.count) : (i += 1) {
            text_w += run.glyphs[i].advance.x;
        }

        return Text{
            .mode = .{ .run = run },
            .width = text_w,
        };
    } else {
        const glyphs = try ally.alloc(*const fcft.Glyph, codepoints.len);
        errdefer ally.free(glyphs);
        const kerns = try ally.alloc(c_long, codepoints.len);
        errdefer ally.free(kerns);

        var text_w: i32 = 0;
        var i: usize = 0;
        while (i < codepoints.len) : (i += 1) {
            glyphs[i] = try font.rasterizeCharUtf32(codepoints[i], .none);

            kerns[i] = 0;
            if (i > 0) {
                var x_kern: c_long = 0;
                if (font.kerning(codepoints[i - 1], codepoints[i], &x_kern, null)) {
                    kerns[i] = x_kern;
                }
            }

            text_w += @intCast(kerns[i] + glyphs[i].advance.x);
        }

        return Text{
            .mode = .{ .glyphs = .{
                .glyphs = glyphs,
                .kerns = kerns,
            } },
            .width = text_w,
        };
    }
}

pub fn deinit(text: *Text, ally: mem.Allocator) void {
    switch (text.mode) {
        .run => text.mode.run.destroy(),
        .glyphs => {
            ally.free(text.mode.glyphs.glyphs);
            ally.free(text.mode.glyphs.kerns);
        },
    }
}

pub fn render_glyphs(
    text: Text,
    image: *pixman.Image,
    fg_color: []const u8,
    x: i32,
    y: i32,
) !void {
    const fg = try decorations.parse_rgba(fg_color);
    const glyphs = switch (text.mode) {
        .run => text.mode.run.glyphs[0..text.mode.run.count],
        .glyphs => text.mode.glyphs.glyphs,
    };
    var tmp_x = x;

    var i: usize = 0;
    while (i < glyphs.len) : (i += 1) {
        if (text.mode == .glyphs) tmp_x += @intCast(text.mode.glyphs.kerns[i]);

        const c = pixman.Image.createSolidFill(&fg).?;
        defer _ = c.unref();

        switch (pixman.Image.getFormat(glyphs[i].pix)) {
            // Glyph is a pre-rendered image. (e.g. a color emoji)
            .a8r8g8b8 => {
                pixman.Image.composite32(
                    .over,
                    glyphs[i].pix,
                    null,
                    image,
                    0,
                    0,
                    0,
                    0,
                    tmp_x + @as(i32, @intCast(glyphs[i].x)),
                    y - @as(i32, @intCast(glyphs[i].y)),
                    glyphs[i].width,
                    glyphs[i].height,
                );
            },
            // Glyph is an alpha mask.
            else => {
                pixman.Image.composite32(
                    .over,
                    c,
                    glyphs[i].pix,
                    image,
                    0,
                    0,
                    0,
                    0,
                    tmp_x + @as(i32, @intCast(glyphs[i].x)),
                    y - @as(i32, @intCast(glyphs[i].y)),
                    glyphs[i].width,
                    glyphs[i].height,
                );
            },
        }

        tmp_x += glyphs[i].advance.x;
    }
}

pub fn get_fcft_fonts(fonts: []const u8) !*fcft.Font {
    var arena = std.heap.ArenaAllocator.init(ctx.gpa);
    defer arena.deinit();
    const ally = arena.allocator();

    var list = try std.ArrayListUnmanaged([*:0]const u8).initCapacity(ally, 2);
    errdefer list.deinit(ally);

    var it = mem.tokenize(u8, fonts, ",");
    while (it.next()) |font| {
        try list.append(ally, try ally.dupeZ(u8, font));
    }

    const fcft_fonts = try fcft.Font.fromName(list.items[0..], null);
    errdefer fcft_fonts.destroy();

    return fcft_fonts;
}
