// This file is part of zelbar
//
// Copyright (C) 2022 Hugo Machet
//
// This program is free software: you can redistribute tokenizerand/or modify
// tokenizerunder the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3
//
// This program is distributed in the hope that tokenizerwill be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

const Tokenizer = @This();

pub const Error = enum {
    /// '{" in text block
    invalid_text,
    /// Attribute doesn't exists
    unknown_attribute,
    /// Attribute not starting with '%{'
    invalid_attribute,
    /// Attribute not ending with '}'
    attribute_not_closed,
};

pub const Tag = enum {
    /// An error is returned in Tokenizer.err
    invalid,

    /// Content, everything allowed but '{'
    text,

    /// Background color
    /// %{B:0xRRGGBB[AA]}
    attr_background,
    /// Foreground color
    /// %{F:0xRRGGBB[AA]}
    attr_foreground,
    /// Swap background and foreground colors
    /// %{R}
    attr_reverse,

    /// Underline
    /// %{U[:0xRRGGBB[AA]:size]}
    attr_line_u,
    /// Overline
    /// %{O[:0xRRGGBB[AA]:size]}
    attr_line_o,

    /// Left align the text in the bar
    /// %{l}
    attr_align_l,
    /// Center align the text in the bar
    /// %{c}
    attr_align_c,
    /// Right align the text in the bar
    /// %{r}
    attr_align_r,

    /// Offset in the alignment direction
    /// %{X:size}
    attr_offset,

    /// Action on mouse click
    /// %{A:action}
    attr_action,

    /// End of line
    eol,
};

const Token = struct {
    tag: Tag,
    bytes: []const u8,
};

const State = enum {
    start,

    text,

    attr,
    attr_start,
    attr_background,
    attr_foreground,
    attr_reverse,
    attr_line_u,
    attr_line_o,
    attr_align_l,
    attr_align_c,
    attr_align_r,
    attr_offset,
    attr_action,
};

state: State = .start,
err: Error = undefined,

buffer: []const u8,
index: usize = 0,
column: usize = 0,

pub fn init(buffer: []const u8) Tokenizer {
    return Tokenizer{ .buffer = buffer };
}

pub fn next(self: *Tokenizer) Token {
    var tok_start: usize = undefined;

    while (self.index < self.buffer.len) {
        const c = self.buffer[self.index];

        switch (self.state) {
            .start => switch (c) {
                '\n', '\r', '\t' => break,
                '%' => self.state = .attr_start,
                '{' => return self.emit_error(.invalid_attribute),
                else => {
                    tok_start = self.index;
                    self.state = .text;
                },
            },

            .attr_start => switch (c) {
                '{' => self.state = .attr,
                else => return self.emit_error(.invalid_attribute),
            },

            .attr => switch (c) {
                'B' => {
                    tok_start = self.index;
                    self.state = .attr_background;
                },
                'F' => {
                    tok_start = self.index;
                    self.state = .attr_foreground;
                },
                'R' => {
                    tok_start = self.index;
                    self.state = .attr_reverse;
                },
                'U' => {
                    tok_start = self.index;
                    self.state = .attr_line_u;
                },
                'O' => {
                    tok_start = self.index;
                    self.state = .attr_line_o;
                },
                'l' => {
                    tok_start = self.index;
                    self.state = .attr_align_l;
                },
                'c' => {
                    tok_start = self.index;
                    self.state = .attr_align_c;
                },
                'r' => {
                    tok_start = self.index;
                    self.state = .attr_align_r;
                },
                'X' => {
                    tok_start = self.index;
                    self.state = .attr_offset;
                },
                'A' => {
                    tok_start = self.index;
                    self.state = .attr_action;
                },
                else => return self.emit_error(.unknown_attribute),
            },

            .attr_background => switch (c) {
                'a'...'z', 'A'...'Z', '0'...'9', ':' => {},
                '}' => return self.emit(.start, .{
                    .tag = .attr_background,
                    .bytes = self.buffer[tok_start..self.index],
                }),
                else => return self.emit_error(.attribute_not_closed),
            },
            .attr_foreground => switch (c) {
                'a'...'z', 'A'...'Z', '0'...'9', ':' => {},
                '}' => return self.emit(.start, .{
                    .tag = .attr_foreground,
                    .bytes = self.buffer[tok_start..self.index],
                }),
                else => return self.emit_error(.attribute_not_closed),
            },
            .attr_reverse => switch (c) {
                '}' => return self.emit(.start, .{
                    .tag = .attr_reverse,
                    .bytes = self.buffer[tok_start..self.index],
                }),
                else => return self.emit_error(.attribute_not_closed),
            },
            .attr_line_u => switch (c) {
                'a'...'z', 'A'...'Z', '0'...'9', ':' => {},
                '}' => return self.emit(.start, .{
                    .tag = .attr_line_u,
                    .bytes = self.buffer[tok_start..self.index],
                }),
                else => return self.emit_error(.attribute_not_closed),
            },
            .attr_line_o => switch (c) {
                'a'...'z', 'A'...'Z', '0'...'9', ':' => {},
                '}' => return self.emit(.start, .{
                    .tag = .attr_line_o,
                    .bytes = self.buffer[tok_start..self.index],
                }),
                else => return self.emit_error(.attribute_not_closed),
            },
            .attr_align_l => switch (c) {
                '}' => return self.emit(.start, .{
                    .tag = .attr_align_l,
                    .bytes = self.buffer[tok_start..self.index],
                }),
                else => return self.emit_error(.attribute_not_closed),
            },
            .attr_align_c => switch (c) {
                '}' => return self.emit(.start, .{
                    .tag = .attr_align_c,
                    .bytes = self.buffer[tok_start..self.index],
                }),
                else => return self.emit_error(.attribute_not_closed),
            },
            .attr_align_r => switch (c) {
                '}' => return self.emit(.start, .{
                    .tag = .attr_align_r,
                    .bytes = self.buffer[tok_start..self.index],
                }),
                else => return self.emit_error(.attribute_not_closed),
            },
            .attr_offset => switch (c) {
                '0'...'9', ':' => {},
                '}' => return self.emit(.start, .{
                    .tag = .attr_offset,
                    .bytes = self.buffer[tok_start..self.index],
                }),
                else => return self.emit_error(.attribute_not_closed),
            },
            .attr_action => switch (c) {
                '}' => return self.emit(.start, .{
                    .tag = .attr_action,
                    .bytes = self.buffer[tok_start..self.index],
                }),
                else => {},
            },

            .text => switch (c) {
                '\n', '\r', '\t' => return self.emit(.start, .{
                    .tag = .text,
                    .bytes = self.buffer[tok_start..self.index],
                }),
                '%' => {
                    switch (self.buffer[self.index + 1]) {
                        '\n', '\r', '\t' => return self.emit(.start, .{
                            .tag = .text,
                            .bytes = self.buffer[tok_start .. self.index + 1],
                        }),
                        '{' => return self.emit(.attr_start, .{
                            .tag = .text,
                            .bytes = self.buffer[tok_start..self.index],
                        }),
                        else => {},
                    }
                },
                '{' => return self.emit_error(.invalid_text),
                else => {},
            },
        }

        self.advance();
    }

    return Token{
        .tag = .eol,
        .bytes = self.buffer[self.buffer.len..],
    };
}

fn advance(self: *Tokenizer) void {
    self.index += 1;
    self.column += 1;
}

fn emit(self: *Tokenizer, next_state: State, tok: Token) Token {
    self.state = next_state;
    self.advance();
    return tok;
}

fn emit_error(self: *Tokenizer, err: Error) Token {
    self.err = err;
    return Token{
        .tag = .invalid,
        .bytes = self.buffer[self.index..],
    };
}

test "attributes block tokenizer (good input)" {
    const testing = @import("std").testing;

    {
        const buf: []const u8 = "%{l}%{c}%{r}";
        var tokenizer = Tokenizer.init(buf);

        var tok = tokenizer.next();
        try testing.expect(tok.tag == .attr_align_l);
        try testing.expectEqualStrings(tok.bytes, "l");

        tok = tokenizer.next();
        try testing.expect(tok.tag == .attr_align_c);
        try testing.expectEqualStrings(tok.bytes, "c");

        tok = tokenizer.next();
        try testing.expect(tok.tag == .attr_align_r);
        try testing.expectEqualStrings(tok.bytes, "r");
    }

    {
        const buf: []const u8 = "%{B:0x000000ff}%{F:0xffffff}%{R}";
        var tokenizer = Tokenizer.init(buf);

        var tok = tokenizer.next();
        try testing.expect(tok.tag == .attr_background);
        try testing.expectEqualStrings(tok.bytes, "B:0x000000ff");

        tok = tokenizer.next();
        try testing.expect(tok.tag == .attr_foreground);
        try testing.expectEqualStrings(tok.bytes, "F:0xffffff");

        tok = tokenizer.next();
        try testing.expect(tok.tag == .attr_reverse);
        try testing.expectEqualStrings(tok.bytes, "R");
    }

    {
        const buf: []const u8 = "%{U:0x000000:2}%{O:0xffffff00:4}%{U:0x000000}%{O:0xffffff00}%{U}%{O}";
        var tokenizer = Tokenizer.init(buf);

        var tok = tokenizer.next();
        try testing.expect(tok.tag == .attr_line_u);
        try testing.expectEqualStrings(tok.bytes, "U:0x000000:2");

        tok = tokenizer.next();
        try testing.expect(tok.tag == .attr_line_o);
        try testing.expectEqualStrings(tok.bytes, "O:0xffffff00:4");

        tok = tokenizer.next();
        try testing.expect(tok.tag == .attr_line_u);
        try testing.expectEqualStrings(tok.bytes, "U:0x000000");

        tok = tokenizer.next();
        try testing.expect(tok.tag == .attr_line_o);
        try testing.expectEqualStrings(tok.bytes, "O:0xffffff00");

        tok = tokenizer.next();
        try testing.expect(tok.tag == .attr_line_u);
        try testing.expectEqualStrings(tok.bytes, "U");

        tok = tokenizer.next();
        try testing.expect(tok.tag == .attr_line_o);
        try testing.expectEqualStrings(tok.bytes, "O");
    }

    {
        const buf: []const u8 = "%{X:10}";
        var tokenizer = Tokenizer.init(buf);

        var tok = tokenizer.next();
        try testing.expect(tok.tag == .attr_offset);
        try testing.expectEqualStrings(tok.bytes, "X:10");

        tok = tokenizer.next();
        try testing.expect(tok.tag == .eol);
        try testing.expectEqualStrings(tok.bytes, "");
    }
}

test "tokenizer (good input)" {
    const testing = @import("std").testing;

    const buf: []const u8 = "%{l}%{A:ran-dom cmd}TEXT%{c}%{R}TEXT%TEXT%{B:0x000000}TEXT% TEXT%%{U:0x000000ff:4}TEXT%%%\n";
    var tokenizer = Tokenizer.init(buf);

    var tok = tokenizer.next();
    try testing.expect(tok.tag == .attr_align_l);
    try testing.expectEqualStrings(tok.bytes, "l");

    tok = tokenizer.next();
    try testing.expect(tok.tag == .attr_action);
    try testing.expectEqualStrings(tok.bytes, "A:ran-dom cmd");

    tok = tokenizer.next();
    try testing.expect(tok.tag == .text);
    try testing.expectEqualStrings(tok.bytes, "TEXT");

    tok = tokenizer.next();
    try testing.expect(tok.tag == .attr_align_c);
    try testing.expectEqualStrings(tok.bytes, "c");

    tok = tokenizer.next();
    try testing.expect(tok.tag == .attr_reverse);
    try testing.expectEqualStrings(tok.bytes, "R");

    tok = tokenizer.next();
    try testing.expect(tok.tag == .text);
    try testing.expectEqualStrings(tok.bytes, "TEXT%TEXT");

    tok = tokenizer.next();
    try testing.expect(tok.tag == .attr_background);
    try testing.expectEqualStrings(tok.bytes, "B:0x000000");

    tok = tokenizer.next();
    try testing.expect(tok.tag == .text);
    try testing.expectEqualStrings(tok.bytes, "TEXT% TEXT%");

    tok = tokenizer.next();
    try testing.expect(tok.tag == .attr_line_u);
    try testing.expectEqualStrings(tok.bytes, "U:0x000000ff:4");

    tok = tokenizer.next();
    try testing.expect(tok.tag == .text);
    try testing.expectEqualStrings(tok.bytes, "TEXT%%%");

    tok = tokenizer.next();
    try testing.expect(tok.tag == .eol);
    try testing.expectEqualStrings(tok.bytes, "");
}

test "attributes block tokenizer (bad input)" {
    const testing = @import("std").testing;

    {
        const buf: []const u8 = "%{l\n";
        var tokenizer = Tokenizer.init(buf);

        const tok = tokenizer.next();
        try testing.expect(tok.tag == .invalid);
        try testing.expect(tokenizer.err == .attribute_not_closed);
    }

    {
        const buf: []const u8 = "%{l\n}";
        var tokenizer = Tokenizer.init(buf);

        const tok = tokenizer.next();
        try testing.expect(tok.tag == .invalid);
        try testing.expect(tokenizer.err == .attribute_not_closed);
    }

    {
        const buf: []const u8 = "%{W}\n";
        var tokenizer = Tokenizer.init(buf);

        const tok = tokenizer.next();
        try testing.expect(tok.tag == .invalid);
        try testing.expect(tokenizer.err == .unknown_attribute);
    }

    {
        const buf: []const u8 = "{l}\n";
        var tokenizer = Tokenizer.init(buf);

        const tok = tokenizer.next();
        try testing.expect(tok.tag == .invalid);
        try testing.expect(tokenizer.err == .invalid_attribute);
    }

    {
        const buf: []const u8 = "TEXT {l}\n";
        var tokenizer = Tokenizer.init(buf);

        const tok = tokenizer.next();
        try testing.expect(tok.tag == .invalid);
        try testing.expect(tokenizer.err == .invalid_text);
    }

    {
        const buf: []const u8 = "TEXT% {l}\n";
        var tokenizer = Tokenizer.init(buf);

        const tok = tokenizer.next();
        try testing.expect(tok.tag == .invalid);
        try testing.expect(tokenizer.err == .invalid_text);
    }
}
