// This file is part of zelbar
//
// Copyright (C) 2022 Hugo Machet
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

const std = @import("std");
const fmt = std.fmt;
const mem = std.mem;

const pixman = @import("pixman");

const ctx = &@import("main.zig").context;
const decorations = @import("decorations.zig");
const Text = @import("Text.zig");
const Tokenizer = @import("Tokenizer.zig");

const log = std.log.scoped(.entry);

/// Parse and render entries data.
pub fn expose(input: []const u8, pix: *pixman.Image, bar_x: u16, bar_w: u16) !void {
    // Height an Entry can use.
    const entry_h = blk: {
        if (ctx.bar.config.border.bottom + ctx.bar.config.border.top >= ctx.bar.config.height) {
            break :blk ctx.bar.config.height;
        }
        break :blk ctx.bar.config.height - ctx.bar.config.border.bottom - ctx.bar.config.border.top;
    };

    var entries = try parse_input(input, entry_h);
    defer entries.deinit();

    const L = 0;
    const C = 1;
    const R = 2;

    const entries_w = entries.get_total_width();

    // We start each alignment from their rightmost point. Use 3 variables to
    // works with every possible order in the raw input.
    var entries_x: [3]u16 = .{
        bar_x + ctx.bar.config.border.left + entries_w[L],
        bar_x + bar_w / 2 + entries_w[C] / 2,
        bar_x + bar_w - ctx.bar.config.border.right,
    };

    var index: usize = entries.list.len;
    while (index > 0) {
        index -= 1;

        var e = entries.list.get(index);
        switch (e.alignment) {
            .left => entries_x[L] -= try e.render(pix, entries_x[L], entry_h),
            .center => entries_x[C] -= try e.render(pix, entries_x[C], entry_h),
            .right => entries_x[R] -= try e.render(pix, entries_x[R], entry_h),
        }
    }

    // Check cases where entries would overlap each other.
    if (ctx.bar.req_width) {
        if ((entries_w[L] + entries_w[C] + entries_w[R]) > bar_w) {
            log.err("bar width({d}) smaller than required width({d})", .{
                bar_w,
                entries_w[L] + entries_w[C] + entries_w[R],
            });
        }
        if (entries_w[C] > 0 and (entries_x[L] + entries_w[L]) > entries_x[C]) {
            log.err("bar width({d}) smaller than required width({d})", .{
                bar_w,
                bar_w + (((entries_x[L] + entries_w[L]) - entries_x[C]) << 1),
            });
        }
        if (entries_w[C] > 0 and (entries_x[C] + entries_w[C]) > entries_x[R]) {
            log.err("bar width({d}) smaller than required width({d})", .{
                bar_w,
                bar_w + (((entries_x[C] + entries_w[C]) - entries_x[R]) << 1),
            });
        }
        if (entries_w[C] == 0 and entries_w[R] > 0 and (entries_x[L] + entries_w[L]) > entries_x[R]) {
            log.err("bar width({d}) smaller than required width({d})", .{
                bar_w,
                bar_w + ((entries_x[L] + entries_w[L]) - entries_x[R]),
            });
        }
    }
}

fn parse_input(input: []const u8, height: u16) !Entries {
    var entries = Entries.new();
    var entry = Entry.new();

    const ally = entries.arena.allocator();

    var tokenizer = Tokenizer.init(input[0..]);
    while (true) {
        const token = tokenizer.next();
        switch (token.tag) {
            .text => {
                entry.text = try Text.new(ally, token.bytes, ctx.bar.font);
                entry.width = @intCast(entry.text.width);

                try entries.list.append(ally, entry);

                if (tokenizer.index < input.len - 1) entry = Entry.new();
            },

            .attr_align_l => entry.alignment = .left,
            .attr_align_c => entry.alignment = .center,
            .attr_align_r => entry.alignment = .right,

            .attr_background => try entry.colors.putNoClobber(ally, .bg, token.bytes[2..]),
            .attr_foreground => try entry.colors.putNoClobber(ally, .fg, token.bytes[2..]),
            .attr_reverse => {
                const fg = entry.colors.get(.fg) orelse ctx.bar.config.colors.get(.fg).?;
                const old_bg = try entry.colors.fetchPut(ally, .bg, fg);
                if (old_bg) |bg| {
                    try entry.colors.put(ally, .fg, bg.value);
                } else {
                    try entry.colors.put(ally, .fg, ctx.bar.config.colors.get(.bg).?);
                }
            },

            .attr_line_o => {
                const line = try Entry.parse_line(height, token.bytes, .over);
                try entry.lines.putNoClobber(ally, .over, line);
            },
            .attr_line_u => {
                const line = try Entry.parse_line(height, token.bytes, .under);
                try entry.lines.putNoClobber(ally, .under, line);
            },

            .attr_offset => entry.offset = try fmt.parseUnsigned(u15, token.bytes[2..], 10),

            .attr_action => entry.action = token.bytes[2..],

            .eol => return entries,

            .invalid => {
                ctx.fatal(.data, "{s} column: {d}\n\t{s}", .{
                    @tagName(tokenizer.err),
                    tokenizer.column,
                    input[tokenizer.index..],
                });
            },
        }
    }
}

pub const Entries = struct {
    arena: std.heap.ArenaAllocator,
    list: std.MultiArrayList(Entry),

    fn new() Entries {
        return Entries{
            .arena = std.heap.ArenaAllocator.init(ctx.gpa),
            .list = .{},
        };
    }

    fn deinit(entries: *Entries) void {
        for (entries.list.items(.text)) |*text| {
            text.deinit(entries.arena.allocator());
        }
        entries.arena.deinit();
    }

    /// Get the total width taken by all entries in each alignment.
    fn get_total_width(entries: Entries) [3]u16 {
        var ret: [3]u16 = .{ 0, 0, 0 };

        for (entries.list.items(.alignment), 0..) |algn, index| {
            const entry_w = entries.list.items(.width)[index] + entries.list.items(.offset)[index];
            switch (algn) {
                .left => ret[0] += entry_w,
                .center => ret[1] += entry_w,
                .right => ret[2] += entry_w,
            }
        }

        return ret;
    }
};

const Entry = struct {
    const Line = struct {
        color: []const u8,
        height: u16,
    };

    /// Alignment in the bar.
    alignment: enum { left, center, right },

    colors: std.AutoHashMapUnmanaged(enum { bg, fg }, []const u8),
    lines: std.AutoHashMapUnmanaged(enum { over, under }, Line),

    /// Offset in alignment direction.
    offset: u15,
    width: u16,

    action: ?[]const u8,

    text: Text = undefined,

    pub fn new() Entry {
        return Entry{
            .width = 0,
            .text = undefined,
            .alignment = .left,
            .colors = .{},
            .lines = .{},
            .offset = 0,
            .action = null,
        };
    }

    /// Render the data.
    ///   prev_x: Previous entry ending position, start of this one.
    ///   height: Usable entry height based on bar height.
    ///
    /// Return the total width of the entry to use as 'x' for the next one.
    pub fn render(entry: Entry, pix: *pixman.Image, prev_x: u16, height: u16) !u16 {
        var x: i16 = @intCast(prev_x);
        switch (entry.alignment) {
            .left, .center => x = x + entry.offset - @as(i16, @intCast(entry.width)) - entry.offset,
            .right => x = x - @as(i16, @intCast(entry.width)) - entry.offset,
        }

        const y: i16 = @intCast(if (ctx.bar.config.location == .bottom) ctx.bar.config.border.bottom else ctx.bar.config.border.top);

        const bg = entry.colors.get(.bg) orelse ctx.bar.config.colors.get(.bg).?;
        try decorations.draw_rectangle(pix, x, y, entry.width, height, bg);

        if (entry.action) |action| {
            try ctx.backend.surface.?.hotspots.append(ctx.gpa, .{
                .action = action,
                .x = @intCast(x),
                .y = @intCast(y),
                .width = entry.width,
                .height = height,
            });
        }

        if (entry.lines.get(.over)) |line_o| {
            try decorations.draw_rectangle(
                pix,
                x,
                y,
                entry.width,
                line_o.height,
                line_o.color,
            );
        }

        if (entry.lines.get(.under)) |line_u| {
            try decorations.draw_rectangle(
                pix,
                x,
                y + @as(i16, @intCast(height - line_u.height)),
                entry.width,
                line_u.height,
                line_u.color,
            );
        }

        // Taken from yambar:
        // <https://codeberg.org/dnkl/yambar/src/branch/master/particles/string.c#L85-L102>
        //
        // This tries to center the font around the bar center, by using
        // the font's ascent+descent as total height, and then removing
        // its descent. This way, the part of the font *above* the
        // baseline is centered.
        //
        // "EEEE" will typically be dead center, with the middle of each character being
        // in the bar's center.
        // "eee" will be slightly below the center.
        // "jjj" will be even further below the center.
        //
        // Finally, if the font's descent is negative, ignore it (except
        // for the height calculation). This is unfortunately not based on
        // any real facts, but works very well with e.g. the "Awesome 5"
        // font family.
        const baseline = y + @divFloor((height + ctx.bar.font.ascent + ctx.bar.font.descent), 2) -
            if (ctx.bar.font.descent > 0) ctx.bar.font.descent else 0;

        const fg = entry.colors.get(.fg) orelse ctx.bar.config.colors.get(.fg).?;
        try entry.text.render_glyphs(pix, fg, x, baseline);

        return entry.width + entry.offset;
    }

    /// Parse the raw input content of a line attribute.
    ///
    /// Content format: %{[U|O]:0xRRGGBB[AA]:height} or %{[O|U]:0xRRGGBB[AA]} or %{[0|U]}.
    /// height without a color before is NOT allowed.
    pub fn parse_line(
        entry_height: u16,
        content: []const u8,
        style: enum { over, under },
    ) !Line {
        var parsed: [3]?[]const u8 = undefined;
        var it = mem.tokenize(u8, content, ":");
        var i: usize = 0;
        while (it.next()) |tok| : (i += 1) {
            parsed[i] = tok;
        }

        const height = blk: {
            if (parsed[2]) |h| break :blk try fmt.parseUnsigned(u16, h, 10);
            break :blk ctx.bar.config.line_height;
        };
        const line_h = blk: {
            if (height > entry_height) {
                log.warn(
                    "line height({d}) greater than possible height({d}), setting it to 2",
                    .{ height, entry_height },
                );
                break :blk 2;
            }
            break :blk height;
        };

        return switch (style) {
            .over => Line{
                .color = parsed[1] orelse ctx.bar.config.colors.get(.line).?,
                .height = line_h,
            },
            .under => Line{
                .color = parsed[1] orelse ctx.bar.config.colors.get(.line).?,
                .height = line_h,
            },
        };
    }
};
