// This file is part of zelbar
//
// Copyright (C) 2022 Hugo Machet
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

const std = @import("std");
const fmt = std.fmt;
const math = std.math;

const pixman = @import("pixman");

const ctx = &@import("main.zig").context;

const log = std.log.scoped(.decorations);

/// Rectangle with a colored background and no borders.
pub fn draw_rectangle(
    image: *pixman.Image,
    x: i16,
    y: i16,
    width: u16,
    height: u16,
    color: []const u8,
) !void {
    const c = try parse_rgba(color);

    _ = pixman.Image.fillRectangles(
        .src,
        image,
        &c,
        1,
        &[1]pixman.Rectangle16{
            .{ .x = x, .y = y, .width = width, .height = height },
        },
    );
}

/// Borders inside a rectangle.
pub const Border = struct {
    top: u16 = 0,
    right: u16 = 0,
    bottom: u16 = 0,
    left: u16 = 0,
    color: []const u8 = "0x646e73",

    pub fn draw(
        border: Border,
        image: *pixman.Image,
        x: i16,
        y: i16,
        width: u16,
        height: u16,
    ) !void {
        const c = try parse_rgba(border.color);

        const borders = [4]pixman.Rectangle16{
            // Top
            .{
                .x = x + @as(i16, @intCast(border.left)),
                .y = 0,
                .width = width - border.left - border.right,
                .height = border.top,
            },
            // Bottom
            .{
                .x = x + @as(i16, @intCast(border.left)),
                .y = @intCast(height - border.bottom),
                .width = width - border.left - border.right,
                .height = border.bottom,
            },
            // Left
            .{
                .x = x,
                .y = y,
                .width = border.left,
                .height = height,
            },
            // Right
            .{
                .x = x + @as(i16, @intCast(width - border.right)),
                .y = y,
                .width = border.right,
                .height = height,
            },
        };

        _ = pixman.Image.fillRectangles(.src, image, &c, 4, &borders);
    }

    /// Check if the size of the borders would not be greater than the one
    /// of the bar.
    pub fn check_size(border: *Border, width: u31, height: u31) !void {
        if (border.top > height or border.bottom > height or
            border.top + border.bottom > height)
        {
            return error.BordersGreaterHeight;
        }

        if (border.left > width or border.right > width or
            border.left + border.right > width)
        {
            return error.BordersGreaterWidth;
        }
    }
};

/// Parse a color in the format "0xRRGGBB" or "0xRRGGBBAA" and convert it to a
/// pixman.Color.
pub fn parse_rgba(s: []const u8) !pixman.Color {
    check_rgba(.decorations, s);

    var color = try fmt.parseUnsigned(u32, s[2..], 16);
    if (s.len == 8) {
        color <<= 8;
        color |= 0xff;
    }

    const bytes = @as([4]u8, @bitCast(color));

    const r: u16 = bytes[3];
    const g: u16 = bytes[2];
    const b: u16 = bytes[1];
    const a: u16 = bytes[0];

    return pixman.Color{
        .red = @as(u16, r << math.log2(0x101)) + r,
        .green = @as(u16, g << math.log2(0x101)) + g,
        .blue = @as(u16, b << math.log2(0x101)) + b,
        .alpha = @as(u16, a << math.log2(0x101)) + a,
    };
}

/// Check if the color use the right format: "0xRRGGBB" or "0xRRGGBBAA".
pub fn check_rgba(comptime scope: @Type(.EnumLiteral), s: []const u8) void {
    if (s.len != 8 and s.len != 10) {
        ctx.fatal(scope, "color must be '0xRRGGBB' or '0xRRGGBBAA': {s}", .{s});
    }
    if (s[0] != '0' or s[1] != 'x') {
        ctx.fatal(scope, "color must be '0xRRGGBB' or '0xRRGGBBAA': {s}", .{s});
    }
}
