// This file is part of zelbar
//
// Copyright (C) 2022 Hugo Machet
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

const std = @import("std");
const mem = std.mem;
const posix = std.posix;

const wayland = @import("wayland");
const wl = wayland.client.wl;

const ctx = &@import("main.zig").context;

const log = std.log.scoped(.event_loop);

pub fn run() !void {
    const poll_stdin = 0;
    const poll_wayland = 1;

    var pollfds: [2]posix.pollfd = undefined;

    pollfds[poll_stdin] = .{
        .fd = posix.STDIN_FILENO,
        .events = posix.POLL.IN,
        .revents = 0,
    };
    pollfds[poll_wayland] = .{
        .fd = ctx.backend.display.getFd(),
        .events = posix.POLL.IN,
        .revents = 0,
    };

    try posix.sigaction(posix.SIG.TERM, &posix.Sigaction{
        .handler = .{ .handler = handle_sig },
        .mask = posix.empty_sigset,
        .flags = 0,
    }, null);
    try posix.sigaction(posix.SIG.INT, &posix.Sigaction{
        .handler = .{ .handler = handle_sig },
        .mask = posix.empty_sigset,
        .flags = 0,
    }, null);

    while (ctx.initialized) {
        ctx.backend.flush_wayland_and_prepare_read();

        _ = posix.poll(&pollfds, -1) catch |err| {
            ctx.fatal(.event_loop, "failed to poll {s}", .{@errorName(err)});
        };

        if (pollfds[poll_stdin].revents & posix.POLL.ERR != 0) {
            ctx.fatal(.event_loop, "failed to poll", .{});
        } else if (pollfds[poll_stdin].revents & posix.POLL.HUP != 0) {
            continue;
        } else if (pollfds[poll_stdin].revents & posix.POLL.IN != 0) {
            var buf: [4096]u8 = undefined;
            const read = try posix.read(posix.STDIN_FILENO, buf[0..]);
            if (read == 0) break;

            if (!mem.eql(u8, buf[0..read], ctx.bar.input.items[0..])) {
                log.debug("read {d} bytes: {s}", .{ read, buf[0..read] });

                ctx.bar.input.clearRetainingCapacity();
                try ctx.bar.input.appendSlice(ctx.gpa, buf[0..read]);
                ctx.backend.surface.?.set_dirty();
            }
        }

        if (pollfds[poll_wayland].revents & posix.POLL.IN != 0) {
            const errno = ctx.backend.display.readEvents();
            if (errno != .SUCCESS) {
                ctx.fatal(.event_loop, "error reading wayland events: {s}", .{@tagName(errno)});
            }
        } else {
            ctx.backend.display.cancelRead();
        }
    }

    ctx.backend.flush_wayland_and_prepare_read();
}

fn handle_sig(_: c_int) callconv(.C) void {
    ctx.initialized = false;
}
