// This file is part of zelbar
//
// Copyright (C) 2022 Hugo Machet
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

const std = @import("std");
const fmt = std.fmt;
const io = std.io;
const mem = std.mem;
const posix = std.posix;

const build_options = @import("build_options");
const fcft = @import("fcft");

const Backend = @import("Backend.zig");
const Bar = @import("Bar.zig");
const decorations = @import("decorations.zig");
const flags = @import("flags");
const event_loop = @import("event_loop.zig");

const log = std.log.scoped(.zelbar);

const usage =
    \\Usage: zelbar [OPTIONS...]
    \\
    \\
    \\  -h                                   Print this help message and exit
    \\  -version                             Print the version number and exit.
    \\  -o <name>                            Select output by name
    \\  -btm                                 Put the bar at the bottom of the output
    \\  -L [0|1|2]                           Set layer_surface layer.
    \\  -g <width>:<height>                  Geometry of the bar (include borders)
    \\  -a [0|1|2]                           Alignment of the bar on the output,
    \\                                         only used if width is not 0
    \\  -m <top>:<right>:<bottom>:<left>     Margins of the bar
    \\  -fn <fonts>                          Comma separated list of FontConfig
    \\                                         formatted font specifications
    \\  -B <color>                           Background color (0xRRGGBB|0xRRGGBBAA)
    \\  -F <color>                           Foreground color (0xRRGGBB|0xRRGGBBAA)
    \\  -bs <top>:<right>:<bottom>:<left>    Borders size
    \\  -bc <color>                          Borders color (0xRRGGBB|0xRRGGBBAA)
    \\  -lh <pixel>                          [Over|Under]line height
    \\  -lc <color>                          [Over|Under]line color (0xRRGGBB|0xRRGGBBAA)
    \\
;

const Context = struct {
    gpa: mem.Allocator = std.heap.c_allocator,

    backend: Backend = .{},
    bar: Bar = .{},

    initialized: bool = false,

    fn deinit(ctx: *Context) void {
        ctx.bar.deinit();
        ctx.backend.deinit();
    }

    pub fn fatal(
        _: Context,
        comptime scope: @Type(.EnumLiteral),
        comptime format: []const u8,
        args: anytype,
    ) noreturn {
        std.log.scoped(scope).err(format, args);
        posix.exit(1);
    }
};

pub var context: Context = .{};

/// Default colors
const background: []const u8 = "0xf9f9fa";
const foreground: []const u8 = "0x202325";
const line: []const u8 = "0x9e2f59";

pub fn main() anyerror!void {
    // Default configuration changed by cli options.
    var default_cfg: Bar.Config = .{};
    try default_cfg.colors.putNoClobber(context.gpa, .bg, background);
    try default_cfg.colors.putNoClobber(context.gpa, .fg, foreground);
    try default_cfg.colors.putNoClobber(context.gpa, .line, line);

    // Parse cli options.
    const result = flags.parser([*:0]const u8, &.{
        .{ .name = "h", .kind = .boolean },
        .{ .name = "version", .kind = .boolean },
        .{ .name = "o", .kind = .arg },
        .{ .name = "btm", .kind = .boolean },
        .{ .name = "L", .kind = .arg },
        .{ .name = "g", .kind = .arg },
        .{ .name = "a", .kind = .arg },
        .{ .name = "m", .kind = .arg },
        .{ .name = "fn", .kind = .arg },
        .{ .name = "B", .kind = .arg },
        .{ .name = "F", .kind = .arg },
        .{ .name = "bs", .kind = .arg },
        .{ .name = "bc", .kind = .arg },
        .{ .name = "lh", .kind = .arg },
        .{ .name = "lc", .kind = .arg },
    }).parse(std.os.argv[1..]) catch {
        try io.getStdErr().writeAll(usage);
        posix.exit(1);
    };
    if (result.args.len != 0) {
        log.err("unknown option '{s}'", .{result.args[0]});
        io.getStdErr().writeAll(usage) catch {};
        posix.exit(1);
    }

    if (result.flags.h) {
        try io.getStdOut().writeAll(usage);
        posix.exit(0);
    }
    if (result.flags.version) {
        try io.getStdOut().writeAll(build_options.version ++ "\n");
        posix.exit(0);
    }
    if (result.flags.o) |raw| default_cfg.output_name = raw;
    if (result.flags.btm) default_cfg.location = .bottom;
    if (result.flags.L) |raw| {
        const n = try fmt.parseUnsigned(u16, raw, 10);
        switch (n) {
            0 => default_cfg.layer = .bottom,
            1 => default_cfg.layer = .top,
            2 => default_cfg.layer = .overlay,
            else => context.fatal(.config, "unknown layer value: {s}", .{raw}),
        }
    }
    if (result.flags.g) |raw| {
        const arr = try parse_string(2, raw);
        default_cfg.width = arr[0];
        default_cfg.height = arr[1];
    }
    if (result.flags.a) |raw| {
        if (default_cfg.width == 0) {
            context.fatal(.config, "'-a' can only be used when width is not 0", .{});
        }

        const n = try fmt.parseUnsigned(u16, raw, 10);
        switch (n) {
            0 => default_cfg.alignment = .left,
            1 => default_cfg.alignment = .center,
            2 => default_cfg.alignment = .right,
            else => context.fatal(.config, "unknown alignment value: {s}", .{raw}),
        }
    }
    if (result.flags.m) |raw| {
        const arr = try parse_string(4, raw);
        default_cfg.margins = .{
            .top = arr[0],
            .right = arr[1],
            .bottom = arr[2],
            .left = arr[3],
        };
    }
    if (result.flags.@"fn") |raw| default_cfg.fonts_name = raw;
    if (result.flags.B) |raw| {
        decorations.check_rgba(.main, raw);
        try default_cfg.colors.put(context.gpa, .bg, raw);
    }
    if (result.flags.F) |raw| {
        decorations.check_rgba(.main, raw);
        try default_cfg.colors.put(context.gpa, .fg, raw);
    }
    if (result.flags.bs) |raw| {
        const arr = try parse_string(4, raw);
        default_cfg.border = .{
            .top = arr[0],
            .right = arr[1],
            .bottom = arr[2],
            .left = arr[3],
        };
    }
    if (result.flags.bc) |raw| {
        decorations.check_rgba(.main, raw);
        default_cfg.border.color = raw;
    }
    if (result.flags.lh) |raw| {
        default_cfg.line_height = try fmt.parseUnsigned(u16, raw, 10);
    }
    if (result.flags.lc) |raw| {
        decorations.check_rgba(.main, raw);
        try default_cfg.colors.put(context.gpa, .line, raw);
    }

    if (@import("builtin").mode == .Debug) {
        _ = fcft.init(.auto, true, .debug);
    } else {
        _ = fcft.init(.auto, false, .err);
    }
    defer fcft.fini();

    context.bar = try Bar.new(default_cfg);
    context.initialized = try context.backend.init();
    defer context.deinit();

    try event_loop.run();
}

/// Parse a string with delimiter ':' and return [n]u16.
fn parse_string(comptime n: usize, s: []const u8) ![n]u16 {
    var ret: [n]u16 = undefined;
    var it = mem.tokenize(u8, s, ":");
    var i: usize = 0;
    while (it.next()) |num| : (i += 1) {
        ret[i] = fmt.parseUnsigned(u16, num, 10) catch |err| switch (err) {
            error.Overflow => std.math.maxInt(u16),
            else => return err,
        };
    }

    return ret;
}
